
int construct_inst( classid_t class,   
                    int count,         
                    list_info_t *info );

int lmi_construct_int_AVL( attid_t att,
                           int val,
                           list_info_t *info,
                           int att_op ) ;

int lmi_construct_str_AVL( attid_t att,
                           char *val,
                           list_info_t *info,
                           int att_op);

int lmi_construct_any_AVL( attid_t att,
                           octet_t *val,
                           int size,
                           list_info_t *info,
                           int att_op );

int lmi_construct_pk_AVL( octet_t *val,
                          int size,
                          list_info_t *info,
                          int att_op );

int lmi_construct_zero_AVL( attid_t att,
                            int size,
                            list_info_t *info,
                            int att_op );

allocate_LMI_request( list_info_t *req,
                      int size );

deallocate_LMI_request( list_info_t *req );

skip_avl( list_info_t *req );

skip_header( list_info_t *req );

